/* Copyright (c) 2013 APDM
*
* basic_report.js
*
* Creates data navigation and reporting for all studies.
* Requires that single_sensor_plot.js and uitls.js be loaded before this script.
*/

// JSHint stuff
/* global longitudinalPlot:false, utils:false, gaitCycleChart:false,
          singleSensorPlot:false */

// Padding around plot
// TODO: Determine if we can handle this in CSS
var padding = 60;

var minPlotHeight = 200;

var reportData;

// The navigation tree in the left pane
var navTree;

// Map of all the raw data plots being displayed
var rawDataPlots = d3.map();

var serverInfo = utils.parseURLParams(window.location.href);

d3.json(serverInfo.dataFile, function(error, data) {
    if (error) return console.warn(error);
    console.log("Loading report data file: " + serverInfo.dataFile);
    reportData = data;

    convertDates();
    createStudySubnodes();

    // Create our navigation tree
    navTree = navigationTree(d3.select("#nav"), reportData);

    var div = d3.select("#report");
    studiesNodeReport(div, reportData);
});

function getChildren(node) {
    return (node.children === null ? node._children : node.children);
}


// Convert all the dates in the report data to JS Date objects.
function convertDates() {
    // Dates in study
    reportData.children.forEach( function(study) {
        study.startDate = utils.getDateFromString(study.startDate);
        study.endDate = utils.getDateFromString(study.endDate);
        
        study.children.forEach( function(subject) {
            // Dates in sessions
            subject.children.forEach( function(session) {
                if (session.startDate !== "Not started") {
                    session.startDate = utils.getDateFromString(session.startDate);
                }
                // Dates in trials
                session.children.forEach( function(trial) {
                    trial.date = utils.getDateFromString(trial.date);
                });
            });
        });
    });
}


// Create the "Study Sessions" and "Study Trials" nodes.
function createStudySubnodes() {
    reportData.children.forEach( function(study) {
        var sessions = [];
        var trials = [];
        // Subjects in "all subjects" node
        study.children.forEach( function(subject) {
            subject.children.forEach( function(session) {
                // Create the copy of the session data
                sessions.push(session);
                // Get each of the trials in the sessions
                session.children.forEach( function(trial) {
                    trials.push(trial);
                });
            });
        });
        study.children.push({name: "Trials", type: "TRIALS_LIST",
                             trials: trials});
    });
}


function studiesNodeReport(root, studiesNode) {
    var studies = getChildren(studiesNode);

    var div = replaceReport(root);
    div.append("h2").text("All studies");
    var filtersDiv = div.append("div").attr("id", "filters-div");
    var studiesDiv = div.append("div").attr("class", "report_subsection");

    var cols = [
        [ {text:"Name", span:1}, {text:"Start date", span:1},
          {text:"End date", span:1} ]
    ];

    var matrix = [];
    studies.forEach( function(d) {
        var row = [ {text: d.name, func:function(selection) {
            return selection.text(d.name)
                .attr("class", "internal_link")
                .on("click", function() {
                    navTree.expandNode(d);
                });}},
                    utils.getStringFromDate(d.startDate),
                    utils.getStringFromDate(d.endDate) ];
        matrix.push(row);
    });
    var table = dataTable().tableClass("metrics").theadClass("metrics_head")
        .trOddClass("metrics_odd").id("studies-table")
        .colNames(cols).sortable(true);
    studiesDiv.datum(matrix).call(table);
    tableFilters(filtersDiv, table, studiesDiv);
}


function studyReport(root, study) {
    var subjects = getChildren(study);

    var div = replaceReport(root);
    div.append("h2").text(study.name);
    // Summary table for the study
    var summary = div.append("div").attr("class", "report_subsection");
    var table = summary.append("table")
        .attr("class", "summary");
    appendStudyNameRow(table, study);
    var row = table.append("tr");
    row.append("td").attr("class", "summary bold_font").text("Start date");
    row.append("td").attr("class", "summary")
        .text(utils.getStringFromDate(study.startDate));
    row = table.append("tr");
    row.append("td").attr("class", "summary bold_font").text("Description");
    row.append("td").attr("class", "summary").text(study.description);

    div.append("h2").text("All subjects for study");
    var filtersDiv = div.append("div").attr("id", "filters-div");
    var metrics = div.append("div").attr("class", "report_subsection");

    var cols = [
        [ {text:"Public ID", span:1}, {text:"Year of birth", span:1},
          {text:"Height", span:1}, {text:"Control", span:1} ]
    ];
    var matrix = [];
    subjects.forEach( function(d) {
        if (d.type !== "SESSIONS_LIST" && d.type !== "TRIALS_LIST") {
            var row = [ {text: d.publicId, func: function(selection) {
                return selection.text(d.publicId)
                    .attr("class", "internal_link")
                    .on("click", function() {
                        navTree.expandNode(d);
                    } );}},
                        d.yearOfBirth, d.height, d.control ];
            matrix.push(row);
        }
    });
    table = dataTable().tableClass("metrics").theadClass("metrics_head")
        .trOddClass("metrics_odd").id("subjects-list").colNames(cols)
        .sortable(true);
    metrics.datum(matrix).call(table);
    tableFilters(filtersDiv, table, metrics);
}


function studySubjectReport(root, studySubject) {
    var sessions = getChildren(studySubject);

    var div = replaceReport(root);
    div.append("h2").text(studySubject.name);

    // Summary table for study subject
    var summary = div.append("div").attr("class", "report_subsection");
    var table = summary.append("table")
        .attr("class", "summary");
    appendStudyNameRow(table, studySubject);
    appendStudySubjectRow(table, studySubject);

    var row = table.append("tr").attr("class", "summary");
    row.append("td").attr("class", "summary bold_font").text("Year of birth");
    row.append("td").attr("class", "summary").text(studySubject.yearOfBirth);
    row = table.append("tr").attr("class", "summary");
    row.append("td").attr("class", "summary bold_font").text("Height");
    row.append("td").attr("class", "summary").text(studySubject.height);

    // Create a table of the sessions
    div.append("h2").text("All trials for subject");
    var filtersDiv = div.append("div").attr("id", "filters-div");
    var tblDiv = div.append("div").attr("class", "report_subsection");
    var trials = [];
    sessions.forEach(function(d) {
        if (d.type !== "TRIALS_LIST") {
            // Get all the trials associated with the subject
            trials = trials.concat(getChildren(d));
        }
    });
    trialsTable(filtersDiv, tblDiv, trials);
}


function studyTrialsReport(root, node) {
    var trials = node.trials;

    var div = replaceReport(root);
    div.append("h2").text("All trials");
    var filtersDiv = div.append("div").attr("id", "filters-div");
    var trialList = div.append("div").attr("class", "report_subsection");
    var cols = [
        [{text:"Name", span:1}, {text:"Condition", span:1},
         {text:"Subject ID", span:1}, {text:"Session", span:1},
         {text:"Date", span:1}, {text:"Notes", span:1} ]
    ];
    
    var matrix = trials.map(function(d) {
        return [ {text: d.testDefinition.testName, func: function(selection) {
            return selection
                .on("click", function() {
                    navTree.expandNode(d);
                })
                .attr("class", "internal_link")
                .text(d.testDefinition.testName);}},
                 d.testDefinition.conditionName,
                 d.parent.parent.name,
                 d.parent.name,
                 utils.getStringFromDate(d.date), d.notes
               ];
    });

    var table = dataTable().tableClass("metrics").theadClass("metrics_head")
        .trOddClass("metrics_odd").id("trials-table").colNames(cols)
        .sortable(true);
    trialList.datum(matrix).call(table);
    tableFilters(filtersDiv, table, trialList);
}


function sessionReport(root, session) {
    var trials = getChildren(session);

    var div = replaceReport(root);
    var summary = div.append("div").attr("class", "report_subsection");
    summary.append("h2").text(session.name);

    var table = summary.append("table")
        .attr("class", "summary");
    appendStudyNameRow(table, session);
    appendStudySubjectRow(table, session);

    var row = table.append("tr").attr("class", "summary");
    row.append("td").attr("class", "summary bold_font").text("Session name");
    row.append("td").attr("class", "summary").text(session.name);
    row = table.append("tr").attr("class", "summary");
    row.append("td").attr("class", "summary bold_font").text("Session start date");
    row.append("td").attr("class", "summary").text(session.startDate);
    row = table.append("tr").attr("class", "summary");

    summary.append("h3").text("Session notes");
    summary.append("p").attr("class", "notes")
        .text(session.notes === "" ? "None" : session.notes );

    // Get the trials grouped by the test type
    var trialsByType = {};
    trials.forEach( function(d) {
        if (trialsByType[d.testDefinition.testName]) {
            trialsByType[d.testDefinition.testName].push(d);
        } else {
            trialsByType[d.testDefinition.testName] = [d];
        }
    });
    var trialGroups = [];
    for (var trialType in trialsByType) {
        trialGroups.push( {name: trialType, children: trialsByType[trialType]} );
    }

    var sessionList = div.append("div").attr("class", "report_subsection");
    sessionTrialsTables(sessionList, trialGroups);
    var longPlots = div.append("div").attr("class", "report_subsection");
    trialGroups.forEach( function(d) {
            allGroupsLongitudinalPlots(longPlots, getChildren(d));
    });
}

function sessionTrialsTables(root, trialGroups) {

    root.append("h2").text("Session Trials");
    trialGroups.forEach( function(d) {
        var div = root.append("div").attr("id", d.name + "-trialGroups-table");
        var children = getChildren(d);
        div.append("h3").text(d.name);
        // TODO: Fix this
        var filtersDiv = div.append("div").attr("id", "filters-div");
        trialsTable(filtersDiv, div, children);
    });
}

function trialsTable(filtersDiv, root, trials) {
    var cols = [
        [{text:"Type", span:1}, {text:"Condition", span:1},
         {text:"Date", span:1}, {text:"Notes", span:1} ]
    ];
    
    var matrix = trials.map(function(d) {
        return [ { text: d.testDefinition.testName, func: function(selection) {
            return selection
                .on("click", function() {
                    navTree.expandNode(d);
                })
                .attr("class", "internal_link")
                .text(d.testDefinition.testName);}},
                 d.testDefinition.conditionName,
                 utils.getStringFromDate(d.date), d.notes
               ];
    });

    var table = dataTable().tableClass("metrics").theadClass("metrics_head")
        .trOddClass("metrics_odd").id("trials-table").colNames(cols)
        .sortable(true);
    var tblDiv = root.append("div").attr("id", "session-trials-tables");
    tblDiv.datum(matrix).call(table);
    tableFilters(filtersDiv, table, tblDiv);
}

function singleTrialReport(root, trial) {
    var div = replaceReport(root);
    var testName = trial.testDefinition.testName;
    var summary = div.append("div").attr("class", "report_subsection");
    summary.append("h2").text(testName + " Trial");

    var table = summary.append("table")
        .attr("class", "summary");
 
    var row = table.append("tr").attr("class", "summary");
    row.append("td").attr("class", "summary bold_font").text("Trial condition");
    row.append("td").attr("class", "summary")
        .text(trial.testDefinition.conditionName);
    row = table.append("tr").attr("class", "summary");
    row.append("td").attr("class", "summary bold_font").text("Trial Date");
    row.append("td").attr("class", "summary")
        .text(utils.getStringFromDate(trial.date));

    summary.append("h3").text("Trial notes");
    summary.append("p").attr("class", "notes")
        .text(trial.notes === "" ? "None" : trial.notes );

    // Add the trial specific data
    var metricsDiv = div.append("div").attr("id", "trial-metrics-summary")
        .attr("class", "report_subsection");
    var metricsSummary = new MetricsSummary(metricsDiv, trial);

    // Create gait cycle chart for IWalk trials
    if (testName === "IWalk") {
        var gaitCycleDiv = div.append("div").attr("id", "gait-cycle");
        gaitCycleDiv.append("h2").text("Gait Cycle");
        var gaitChart = gaitCycleChart().width($('#report').width() - (padding * 2));

        // Note we only want to send the "GaitTrack: Lower Limb" measures
        var gaitMeasures = getMetricGroup(getMetricClass(trial.metrics, "Measure"),
                                          "GaitTrack: Lower Limb");
        // If above call failed, might be IWalk and not GaitTrack trial
        if (!gaitMeasures.length) {
            gaitMeasures = getMetricGroup(getMetricClass(trial.metrics, "Measure"),
                                          "Gait: Lower Limb");
        }
        gaitCycleDiv.datum(gaitMeasures).call(gaitChart);
    }

    singleTrialRawDataPlots(div, trial);
}

function replaceReport(root) {
    var old = root.select("#detail-report");
    if (old) old.remove();

    var div = root.append("div")
        .attr("class", "report")
        .attr("id", "detail-report");
    return div;
}

function singleTrialRawDataPlots(root, trial) {
    // Plot width determined by width of the #report div
    var plotWidth = $('#report').width() - (padding * 2);
    var plotHeight = ($(window).height() / 2) - (padding * 2);
    plotHeight = d3.max([plotHeight, minPlotHeight]);

    // Get the event data for the plot
    var defs;
    for (var def in  reportData.metricDefinitions) {
        if (reportData.metricDefinitions[def].testName === trial.testDefinition.testName) {
            defs = reportData.metricDefinitions[def];
            break;
        }
    }

    var plotDiv = root.append("div").attr("class", "raw_data_plots");
    var plots = new RawDataPlots(trial, defs, plotDiv, plotWidth, plotHeight);

}

// Generate a metrics summary table given the metrics to display
function MetricsSummary(div, trial) {
    var defs = getDefinitionsForTrial(trial);
    var metrics = getMetricClass(trial.metrics, "Measure");
    var groups = metricsDefGroups(defs.metrics);

    div.append("h2").text("Metrics summary");
    var allTableDiv = div.append("div").attr("class", "all_metrics_tables");

    // Build tables for each of the metric groups
    for(var group in groups) {
        var groupData = [];
        metrics.forEach( function(metric) {
            if (metric.metricGroup === group) {
                groupData.push(metric);
            }
        });
        if (groupData.length > 0) {
            allTableDiv.append("h3").text( function() {
                return ( group === "" ? "General" : group );
            });
            var table = dataTable().tableClass("metrics").theadClass("metrics_head")
                .trOddClass("metrics_odd");
            var tblDiv = allTableDiv.append("div");
            var metricTblData = formatMetricTableData(defs, group, groupData);
            table.colNames(metricTblData[0]);
            tblDiv.datum(metricTblData[1]).call(table);
        }
    }
}

function getDefinitionsForTrial(trial) {
    for (var def in  reportData.metricDefinitions) {
        if (reportData.metricDefinitions[def].testName === trial.testDefinition.testName) {
            return reportData.metricDefinitions[def];
        }
    }
    return null;
}

function metricsDefGroups(defs) {
    var metricGroups = {};
    defs.forEach( function(d) {
        if(metricGroups[d.metricGroup]) {
            metricGroups[d.metricGroup].push(d.metricName);
        } else {
            metricGroups[d.metricGroup] = [d.metricName];
        }
    });
    return metricGroups;
}

function formatMetricTableData(metricDefinitions, group, metrics) {
    var allMetrics = d3.nest()
        .key(function(d) { return d.metricName; })
        .map(metrics, d3.map);

    // Different column labels based on existence of bilateral data,
    // and if the metrics have a combined mean present.
    var cols = [
        [{text:"Metric", span:1}, {text:"Units", span:1},
         {text:"Mean", span:1}, {text:"StdDev",span:1},
         {text:"Norm mean", span:1}, {text:"Norm stdDev", span:1}]
    ];
    
    // TODO: Will need to revisit how bilateral data is displayed.
    var bilatWithMeanCols = [
        [{text:"Metric", span:1}, {text:"Units", span:1},
         {text:"Mean (Right - Left)", span:1}, {text:"StdDev (Right - Left)",span:1},
         {text:"Norm mean", span:1}, {text:"Norm stdDev", span:1}]
    ];
    var bilatNoMeanCols = [
        [{text:"Metric", span:1}, {text:"Units", span:1},
         {text:"Mean (Left / Right)", span:1}, {text:"StdDev (Left / Right)",span:1},
         {text:"Norm mean (Left / Right)", span:1},
         {text:"Norm stdDev (Left / Right)", span:1}]
    ];

    var rows = [];
    var metricClass = getMetricDefClass(metricDefinitions.metrics, "Measure");
    getMetricDefGroup(metricClass, group).forEach(function(d) {
        var aRow = [];
        var metric;
        if (d.metricType === "BILATERAL") {
            // Need to find the L & R metrics for BILATERAL metric definitions.
            if (allMetrics[d.metricName + " L"] && allMetrics[d.metricName + " R"]) {
                var left = allMetrics[d.metricName + " L"][0];
                var right = allMetrics[d.metricName + " R"][0];
                var lrMeanDiff = formatDigits(right.mean - left.mean);
                var lrSdDiff = formatDigits(right.std - left.std);

                // We do different style columns based on if there is
                // a combined mean for the metric
                if (allMetrics[d.metricName]) {
                    metric = allMetrics[d.metricName];
                    aRow = [d.metricName, d.units,
                            formatDigits(metric[0].mean) + " (" + lrMeanDiff + ")",
                            formatDigits(metric[0].std) +  " (" + lrSdDiff + ")",
                            formatDigits(d.meanNorm), formatDigits(d.sdNorm) ];
                    // Change the column labels
                    cols = bilatWithMeanCols;
                } else {
                    aRow = [ d.metricName, d.units,
                             formatDigits(left.mean) + " / " + formatDigits(right.mean),
                             formatDigits(left.std) + " / " + formatDigits(right.std),
                             formatDigits(d.meanNormL) + " / " + formatDigits(d.meanNormR),
                             formatDigits(d.sdNormL) + " / " + formatDigits(d.sdNormR),
                           ];
                    // Change column labels
                    cols = bilatNoMeanCols;
                }
                rows.push(aRow);
            }
        } else {
            metric = allMetrics[d.metricName];
            if (metric) {
                aRow = [metric[0].metricName, d.units,
                        formatDigits(metric[0].mean), formatDigits(metric[0].std),
                        formatDigits(d.meanNorm), formatDigits(d.sdNorm) ];
                rows.push(aRow);
            }
        }
    });
    return [cols, rows];
}

function formatDigits(value) {
    if (value === "") {
        return "--";
    }
    return value.toPrecision(3);
}

// Get all the metric definitions of a given class. Asssumes that 
function getMetricDefClass(defs, metricCls) {
    var classDefs = [];
    defs.forEach( function(d) {
        if (d.metricClass === metricCls) classDefs.push(d);
    });
    return classDefs;
}

// Get all the metric definitions of a given group
function getMetricDefGroup(defs, group) {
    var groupDefs = [];
    defs.forEach( function(d) {
        if (d.metricGroup === group) groupDefs.push(d);
    });
    return groupDefs;
}

function getMetricClass(metrics, metricCls) {
    var classMetrics = [];
    metrics.forEach(function(d) {
        if (d.metricClass === metricCls) classMetrics.push(d);
    });
    return classMetrics;
}

function getMetricGroup(metrics, group) {
    var groupMetrics = [];
    metrics.forEach(function(d) {
        if (d.metricGroup === group) groupMetrics.push(d);
    });
    return groupMetrics;
}

// Class for creation of raw data plots for a Trial
function RawDataPlots(trial, metricDefs, root, plotWidth, plotHeight) {
    var self = this;
    this.sensorPlots = [];
    this.root = root;
    this.trial = trial;
    this.plotWidth = plotWidth;
    this.plotHeight = plotHeight;
    this.defs = metricDefs;
    this.data = null;

    var title = this.root.append("div");
    title.append("h2").style("display", "inline").text("Raw Data Plots");
    title.append("button").text("view")
        .attr("class", "show_raw_data_btn").on("click", function() {
            self.showPlots(this); });
}

RawDataPlots.prototype.showPlots = function (btn) {
    // Disable the button until the plots have been rendered
    d3.select(btn).attr("disabled", "disabled")
        .on(".click", null);

    var self = this;
    // Get plot data from server, then call the function to create plots.
    // Assume we are only getting the Accel and Gyros for now.
    var senList = "[\"Accelerometers\", \"Gyroscopes\"]";
    var json =
        "{\"displayWidth\": " + (this.plotWidth * 1.5) +
        ",\"sensors\": " + senList + "}";
    // Get new data from the server
    $.post("http://" + serverInfo.serverHost + ":" +
           serverInfo.serverPort + "/static/html/",
           {
               requestType: "mobilitylab",
               options: json,
               dataFile: reportData.options.monitorDataDirectory + self.trial.fileName,
               jsonFile: reportData.options.tempDirectory + self.trial.fileName + ".json",
           },
           function(d) {
               self.data = d;
               self.drawPlots();
               d3.select(btn).text("hide").attr("disabled", null)
                   .on("click", function() {
                       self.hidePlots(this); });
           },
           "json");
};

RawDataPlots.prototype.hidePlots = function(btn) {
    var self = this;
    this.showableDiv.remove();
    d3.select(btn).text("view").on("click", function() { self.showPlots(this); });
};

// Generate the actual plot after the data has been loaded.
RawDataPlots.prototype.drawPlots = function () {
    this.showableDiv = this.root.append("div");
    this.totalHeight = (this.plotHeight + padding) * this.data.monitors[0].Calibrated.length;

    // Use just the first monitor to determine the start time for
    // event time conversion
    var events = formatEventData(this.defs, this.trial.metrics,
                                 this.data.monitors[0].Calibrated[0].data[0].time / 1000);

    for (var i = 0; i < this.data.monitors.length; i++) {
        var monitor = this.data.monitors[i];
        var caseId = monitor.caseId;

        this.id = caseId + "-" + this.trial.fileName.replace(".", "_");
        var plotDiv = this.showableDiv.append("div")
            .attr("id", this.id);

        plotDiv.append("h3")
            .text(monitor.monitorLabel);
        var legend = plotDiv.append("div")
            .attr("id", "plot-key_" + this.id);
        generateLegend(legend);
        generateEventsKey(legend, events);
        var svg = plotDiv.append("svg")
            .attr("class", "single_plot")
            .attr("width", this.plotWidth)
            .attr("height", this.totalHeight);
        for (var k = 0; k < monitor.Calibrated.length; k++) {
            var yOffset;
            if (k === 0) {
                yOffset = 0;
            } else {
                yOffset = (k * padding) + (k * this.plotHeight);
            }
            this.sensorPlots.push(new SingleRawDataPlot(svg,
                                                        this.plotWidth, this.plotHeight,
                                                        0, yOffset,
                                                        monitor.Calibrated[k].name +
                                                        "_" + this.id,
                                                        monitor.Calibrated[k],
                                                        events));
        }
    }
};


function SingleRawDataPlot(svg, plotWidth, plotHeight, xOffset, yOffset,
                           id, sensorData, events) {
    var self = this;
    this.svg = svg;
    this.zoom = d3.behavior.zoom().scaleExtent([1, 100]).on("zoom", self.zoomChanged);
    this.id = "plot-area-" + id;
    var margins = { top: 20, right:20, bottom:40, left:60 };

    this.plotG = this.svg.append("g")
        .attr("id", this.id)
        .attr("transform", "translate(" + xOffset + "," + yOffset + ")");

    this.plot = singleSensorPlot()
        .time1( function(d) {
            return (new Date(0).setUTCMilliseconds(d.time/1000));
        })
        .x1(function(d) { return d.x; })
        .y1(function(d) { return d.y; })
        .z1(function(d) { return d.z; })
        .scalar1(null)
        .width(plotWidth).height(plotHeight)
        .vertLabel(sensorData.name + " (" + sensorData.attributes.Units + ")")
        .horzLabel("Time")
        .margin(margins)
        .eventStart(function(d) { return d.startTime; })
        .eventEnd(function(d) { return d.endTime; })
        .eventType(function(d) { return d.type; })
        .eventID(function(d) { return d.id; })
        .plotEvents(true)
        .displayRelativeTime(true);

    this.plotG.datum([sensorData.data, events]).call(this.plot);

    var pointerG = this.plotG.append("g")
        .attr("transform", "translate(" + margins.left + ",0)");
    pointerG.append("rect")
        .attr("class", "pointer_mask")
        .attr("id", this.id)
        .attr("x", 0)
        .attr("y", margins.top)
        .attr("width", plotWidth - margins.right - margins.left)
        .attr("height", plotHeight - margins.bottom - margins.top)
        .call(this.zoom);

    // We need to keep track of these so that we can get a reference
    // to the SensorPlot when the zoom is called.
    rawDataPlots.set(this.id, this);
}


SingleRawDataPlot.prototype.zoomChanged = function() {
    // The "this" in this function references the DOM element that the
    // zoom behavior is attached to.
    var p = rawDataPlots.get(this.id);
    p.plot.scale(p.zoom.scale()).translate(p.zoom.translate());
    p.plotG.call(p.plot);
};

// Generate a legend for a plot
function generateLegend(div) {

    var len = 40;
    var lines = ["x", "y", "z"];
    var offset = 0;

    var svg = div.attr("class", "plot-key")
        .append("svg").attr("height", 20);
    svg.selectAll("g")
        .data(lines)
        .enter()
        .append("g")
        .each(function(d, i) {
            d3.select(this)
                .attr("transform", function() {
                    if (i !== 0) {
                        offset = ((len + 20) * i);
                    }
                    return "translate(" + offset + ",0)";
                });
            d3.select(this)
                .append("line")
                .attr("class", "line_" + d)
                .attr("x1", 0).attr("y1", 10)
                .attr("x2", len).attr("y2", 10);
            d3.select(this)
                .append("text")
                .attr("text-anchor", "start")
                .attr("x", len + 5)
                .attr("y", 10)
                .text(d);
        });
}

// Create the key for plotted events
function generateEventsKey(root, events) {
    if (events.length < 1) return;

    // Pull out all the unique event types
    var evDict = d3.nest()
        .key(function(d) { return d.type; })
        .map(events, d3.map);
    var evTypes = [];
    var gaitCycleAdded = false;
    for (var type in evDict) {
        // Need to handle the gait cycle differently
        if (type === "gait_cycle_a" || type === "gait_cycle_b") {
            if (!gaitCycleAdded) {
                gaitCycleAdded = true;
                evTypes.push( { name: evDict[type][0].name,
                                type: "gait_cycle" });
            }
        } else {
            evTypes.push( { name: evDict[type][0].name,
                            type: type });
        }
    }
    var spacing = 130;
    var svg = root.append("div").attr("class", "event_key")
        .append("svg")
        .attr("height", 20)
        .attr("width", evTypes.length * spacing);
    
    svg.selectAll("g")
        .data(evTypes)
        .enter()
        .append("g")
        .each(function(d, i) {
            var g = d3.select(this);
            if (d.type === "gait_cycle") {
                g.append("rect")
                    .attr("class", "event_rect " + d.type + "_a")
                    .attr("x",  i * spacing)
                    .attr("y", 0)
                    .attr("width", 10).attr("height", 20);
                g.append("rect")
                    .attr("class", "event_rect " + d.type + "_b")
                    .attr("x", (i * spacing) + 10)
                    .attr("y", 0)
                    .attr("width", 10).attr("height", 20);
            } else {
                g.append("rect")
                    .attr("class", "event_rect " + d.type)
                    .attr("x", i * spacing)
                    .attr("y", 0)
                    .attr("width", 20).attr("height", 20);
            }
        });

    svg.selectAll("g")
        .data(evTypes)
        .append("text")
        .attr("text-anchor", "start")
        .attr("x", function(d, i) { return (i * spacing) + 24; })
        .attr("y", 13)
        .text(function(d) { return d.name; });
}

function appendStudyNameRow(table, data) {
    while (data.type !== "STUDY") {
        data = data.parent;
    }
    var row = table.append("tr").attr("class", "summary");
    row.append("td").attr("class", "summary bold_font").text("Study name");
    row.append("td").attr("class", "summary").text(data.name);
}

function appendStudySubjectRow(table, data) {
    while (data.type !== "STUDY_SUBJECT") {
        data = data.parent;
    }
    var row = table.append("tr").attr("class", "summary");
    row.append("td").attr("class", "summary bold_font").text("Subject ID");
    row.append("td").attr("class", "summary").text(data.name);
}


// Create the table filters, and when used, update the table
function tableFilters(filterDiv, table, tableDiv) {
    var filterValues = [];

    // Get the values for each of the columns out of the table
    var thList = tableDiv.selectAll("th");
    var colNames = [];
    thList.each( function(d) { colNames.push(d.text); });

    colNames.forEach( function(col, i) {
        filterValues[i] = {text: col, values: []};
    });

    // TODO: Determine if there is a better way to get unique values out of this
    // Get the row values out of the tables
    var trList = tableDiv.selectAll("tbody tr");
    trList.each( function() {
        var tdList = d3.select(this).selectAll("td");
        tdList.each( function(d, i) {
            var tdText = d3.select(this).node().innerText ||
                d3.select(this).node().textContent;
            if (filterValues[i].values.indexOf(tdText) <= -1) {
                filterValues[i].values.push(tdText);
            }
        });
    });
    
    var buttonDiv = filterDiv.append("div").attr("id", "filter-buttons");
    buttonDiv.append("span").attr("class", "bold_font one_point_twofive").text("Filters:");
    var clickCb = filterSelectBox(filterDiv, filterValues, table, tableDiv);
    buttonDiv.selectAll("button")
        .data(filterValues).enter()
        .append("button").attr("name", function(d) { return d.text; } )
        .attr("value", function(d, i) { return i; })
        .attr("class", "normal_button")
        .text(function(d) { return d.text; })
        .call(clickCb);
}


function filterSelectBox(root, thefilters, table, tableDiv) {
    var boxOpen = false;
    var filterBoxDiv = null;
    var filterBox = null;
    var filterValues = thefilters;
    var curFilterCol = 0;
    var tableData = table.rowData();
    var selectedFilters = [];

    function setHandler(selection) {
        selection.on("click", updateFilterBox);
    }

    function updateFilterBox() {
        if (!boxOpen) {
            curFilterCol = parseInt(this.value, 10);
            buildFilterBox();
        } else if (boxOpen && curFilterCol !== parseInt(this.value, 10)) {
            filterBoxDiv.remove();
            curFilterCol = parseInt(this.value, 10);
            buildFilterBox();
        } else {
            boxOpen = false;
            filterBoxDiv.remove();
        }
    }

    function buildFilterBox() {
        boxOpen = true;
        filterBoxDiv = root.append("div").attr("id", "transfer-box")
            .style("width", "500px");
        var height = filterValues[curFilterCol].length > 10 ? 10 : filterValues[curFilterCol].length;
        filterBox = $("#transfer-box")
            .bootstrapTransfer({'target_id': 'box-select',
                                'height': height + 'em'});
        $("#box-select").on("transfer_selection_change", updateTable);
        var values = [];
        filterValues[curFilterCol].values.forEach(
            function(entry, i) {
                values.push( {value:i, content:entry} );
            });
        filterBox.populate(values);
        if (selectedFilters[curFilterCol])
            filterBox.set_values(selectedFilters[curFilterCol]);
    }

    function updateTable() {
        // Add the current selected filters to our list
        var filteredData = tableData.slice(0); // Clone array

        selectedFilters[curFilterCol] = filterBox.get_values();

        // Remove all the table rows that do not match our filters
        selectedFilters.forEach( function(d, i) {
            var filters = [];
            d.forEach( function(value) {
                filters.push(filterValues[i].values[parseInt(value, 10)]);
            });
            filteredData = filterCol(filteredData, i, filters);
        });
        table.rowData(filteredData);
        tableDiv.datum(filteredData).call(table);
    }

    function filterCol(tblRows, colNum, filters) {
        var filterResults = [];
        if (!filters.length)
            return tblRows;

        filters.forEach( function(filter) {
            tblRows.forEach( function(row) {
                var text;
                if (row[colNum] instanceof Object)
                    text = row[colNum].text;
                else
                    text = row[colNum];
                if (text === filter) {
                    filterResults.push(row);
                }
            });
        });
        return filterResults;
    }

    return setHandler;
}

// Reusable function to build table.
function dataTable() {
    var colNameMatrix = [];
    var rowMatrix = [];
    var trOddCl = null;
    var trEvenCl = null;
    var tdCl = null;
    var theadCl = null;
    var trHeadCl = null;
    var tbodyCl = null;
    var tableCl = null;
    var sortable = false;
    var sortedCol = 0;
    var initialSort = true;
    var id = 0;

    function tbl(selection) {
        selection.each( function(tableData, i) {
            rowMatrix = tableData;
            var tableDiv = selection.selectAll("div").data([rowMatrix]);
            var tableEnter = tableDiv.enter().append("div")
                .append("table").attr("id", id)
                .attr("class", tableCl);
            if (sortable)
                tableEnter.attr("class", "tablesorter " + tableCl);

            tableEnter.append("thead").attr("class", theadCl);
            tableEnter.append("tbody").attr("class", tbodyCl);

            // Update the table if needed
            tableEnter = tableDiv.select("table")
                .attr("class", "tablesorter " + tableCl)
                .attr("id", id);

            var thead = tableEnter.selectAll("thead");
            var theadRows = thead.selectAll("tr")
                .data(colNameMatrix)
                .enter()
                .append("tr").attr("class", trHeadCl);
            theadRows.selectAll("th")
                .data(function(d) { return d; })
                .enter()
                .append("th")
                .each( function (d,i,k) {
                    var th = d3.select(this);
                    th.attr("class", tdCl)
                        .attr("colspan", function(d) { return d.span; })
                        .text(function(d) { return d.text; });
                    if (sortable && k === colNameMatrix.length - 1) {
                        var cb = new ColClick(th, selection, i);
                        th.node().colClick = cb;
                        th.attr("class", "header " + trHeadCl)
                            .on("click", function() { return cb.onClick(); });
                    }
                });

            var tbody = tableEnter.selectAll("tbody");
            var rows = tbody.selectAll("tr")
                .data(rowMatrix, function(d) {
                    return d.id || (d.id = ++i); });
            var rowsEnter = rows.enter().append("tr");

            rowsEnter.selectAll("td")
                .data(function(d) { return d; })
                .enter()
                .append("td").attr("class", tdCl)
                .each( function(d) {
                    if (d instanceof Function) {
                        return d(d3.select(this));
                    } else if (d instanceof Object) {
                        return d.func(d3.select(this));
                    }
                    d3.select(this).text(function(d) { return d; });
                });

            rows.exit().remove();

            // Set the class for the odd and even rows
            rows.each( function(d, i) {
                if (i % 2 === 0) {
                    d3.select(this).attr("class", trEvenCl);
                } else {
                    d3.select(this).attr("class", trOddCl);
                }
            });
            // Sort the rows on the currently sorted column
            if (sortable === true) {
            thead.selectAll("th")
                .each( function(d, i) {
                    if (i === sortedCol) {
                        var cb = d3.select(this).node().colClick;
                        if (initialSort === false) {
                            // Flip the flag if this is NOT the first time through
                            if (cb.asending === true) {
                                cb.asending = false;
                            } else {
                                cb.asending = true;
                            }
                        }
                        initialSort = false;
                        cb.onClick();
                    }
                });
            }
        });
    }

    function ColClick(th, sel, idx) {
        this.asending = false;
        this.th = th;
        this.sel = sel;
        this.idx = idx;

        this.onClick = function() {
            // Sort the matrix based on the column index
            sortedCol = this.idx;
            var rows = this.sel.select("tbody").selectAll("tr");
            var ths = this.sel.select("thead").selectAll("th");

            if (this.asending) {
                this.asending = false;
                ths.attr("class", "header");
                this.th.attr("class", "header headerSortUp");
                rows.sort(function(a, b) { return rowsAcending(a[idx], b[idx]); });
            } else {
                this.asending = true;
                ths.attr("class", "header");
                this.th.attr("class", "header headerSortDown");
                rows.sort(function(a, b) { return rowsDecending(a[idx], b[idx]); });
            }
            rows.each( function(d, i) {
                if (i % 2 === 0) {
                    d3.select(this).attr("class", trEvenCl);
                } else {
                    d3.select(this).attr("class", trOddCl);
                }
            });
        };
    }

    function rowsDecending(a, b) {
        if (a instanceof Object && b instanceof Object) {
            a = a.text;
            b = b.text;
        }
        a = a.toLowerCase();
        b = b.toLowerCase();
        return a > b ? 1 : a === b ? 0 : -1;
    }
    

    function rowsAcending(a, b) {
        if (a instanceof Object && b instanceof Object) {
            a = a.text;
            b = b.text;
        }
        a = a.toLowerCase();
        b = b.toLowerCase();
        return a < b ? 1 : a === b ? 0 : -1;
    }


    tbl.colNames = function(d) {
        if (!arguments.length) return colNameMatrix;
        colNameMatrix = d;
        return tbl;
    };

    tbl.rowData = function(d) {
        if (!arguments.length) return rowMatrix;
        rowMatrix = d;
        return tbl;
    };

    tbl.sortable = function(d) {
        if (!arguments.length) return sortable;
        sortable = d;
        return tbl;
    };

    tbl.sortedCol = function(d) {
        if (!arguments.length) return sortedCol;
        sortedCol = d;
        return tbl;
    };

    tbl.trOddClass = function(d) {
        if (!arguments.length) return trOddCl;
        trOddCl = d;
        return tbl;
    };

    tbl.trEvenClass = function(d) {
        if (!arguments.length) return trEvenCl;
        trEvenCl = d;
        return tbl;
    };

    tbl.tdClass = function(d) {
        if (!arguments.length) return tdCl;
        tdCl = d;
        return tbl;
    };
    tbl.theadClass = function(d) {
        if (!arguments.length) return theadCl;
        theadCl = d;
        return tbl;
    };
    tbl.trClass = function(d) {
        if (!arguments.length) return tbodyCl;
        tbodyCl = d;
        return tbl;
    };
    tbl.tableClass = function(d) {
        if (!arguments.length) return tableCl;
        tableCl = d;
        return tbl;
    };

    tbl.id = function(d) {
        if (!arguments.length) return id;
        id = d;
        return tbl;
    };

    return tbl;
}


function formatEventData(metricDefinitions, metrics, startTime) {
    var eventDefs = getMetricDefClass(metricDefinitions.metrics, "Event");
    var eventMetrics = getMetricClass(metrics, "Event");

    var sitToStandEvents = getSitToStandEvents(eventDefs, eventMetrics, startTime);
    var gaitCycleEvents = getGaitCycleEvents(eventDefs, eventMetrics, startTime);
    var turnEvents = getTurnEvents(eventDefs, eventMetrics, startTime);
    var turnToSitEvents = getTurnToSitEvents(eventDefs, eventMetrics, startTime);

    return sitToStandEvents.concat(gaitCycleEvents, turnEvents, turnToSitEvents);
}


// Process the Sit-to-Stand events into an array of objects with start
// and stop time
function getSitToStandEvents(eventDefs, allEvents, startTime) {
    var events = getMetricGroup(allEvents, "Postural Transitions: Sit to Stand");
    var startEndTimes = {};
    var sitToStand = [];

    startEndTimes = getEventStartEndTimes(events, startTime);
    if (!startEndTimes) return [];

    startEndTimes.startTimes.forEach( function(d, i) {
        sitToStand.push( { name: "Sit to Stand",
                           id: "sit-to-stand",
                           type: "sit_to_stand",
                           startTime: d,
                           endTime: startEndTimes.endTimes[i],
                         });
    });

    return sitToStand;
}


// Process the gait events into an array of objects with simple start
// and stop times for each gait cycle.
function getGaitCycleEvents(eventDefs, allEvents, startTime) {
    var events = getMetricGroup(allEvents, "Gait: Lower Limb");
    var gaitCycles = [];
    var startTimes = [];
    var endTimes = [];

    events.forEach( function(d) {
        if (d.metricName === "Initial Contact Right") startTimes = d.values;
        if (d.metricName === "Next Initial Contact Right") endTimes = d.values;
    });

    // Build the list of objects, if we can
    if ((!startTimes || !endTimes) || (!startTimes.length || !endTimes.length))
        return [];

    // Convert the times to UTC milliseconds so plot can use them
    startTimes = startTimes.map(function(d) {
        return (d * 1000) + startTime;
    });

    endTimes = endTimes.map(function(d) {
        return (d * 1000) + startTime;
    });

    startTimes.forEach( function(d, i) {
        // Need to assign different classes to the gait cycles, since
        // they can be right next to each other and therefore need
        // different colors.
        var gaitType;
        if (i % 2 === 0)
            gaitType = "gait_cycle_a";
        else
            gaitType =  "gait_cycle_b";
        gaitCycles.push( { name: "Gait cycles",
                           id: "gait-cycle",
                           type: gaitType,
                           startTime: d,
                           endTime: endTimes[i],
                         });
    });
    
    return gaitCycles;
}

function getTurnEvents(eventDefs, allEvents, startTime) {
    var events = getMetricGroup(allEvents, "Postural Transitions: Turning");
    var startEndTimes = {};
    var turns = [];

    startEndTimes = getEventStartEndTimes(events, startTime);
    if (!startEndTimes) return [];

    startEndTimes.startTimes.forEach( function(d, i) {
        turns.push( { name: "Turn", id: "turn", type: "turn",
                           startTime: d, endTime: startEndTimes.endTimes[i],
                         });
    });

    return turns;
}

function getTurnToSitEvents(eventDefs, allEvents, startTime) {
    var events = getMetricGroup(allEvents, "Postural Transitions: Turn to Sit");
    var startEndTimes = {};
    var turnToSits = [];

    startEndTimes = getEventStartEndTimes(events, startTime);
    if (!startEndTimes) return [];

    startEndTimes.startTimes.forEach( function(d, i) {
        turnToSits.push( { name: "Turn to sit", id: "turn-to-sit",
                           type: "turn_to_sit",
                           startTime: d, endTime: startEndTimes.endTimes[i],
                         });
    });

    return turnToSits;
}


function getEventStartEndTimes(events, timeOffset) {
    // For this event there should be two events, start and stop,
    // where each event contains an array of start and stop times
    var startTimes = [];
    var endTimes = [];
    events.forEach( function(d) {
        if (d.metricName === "Start Time") startTimes = d.values;
        if (d.metricName === "End Time") endTimes = d.values;
    });

    // Build the list of objects, if we can
    if ((!startTimes || !endTimes) || (!startTimes.length || !endTimes.length))
        return null;

    // Convert the times to UTC milliseconds so plot can use them
    startTimes = startTimes.map(function(d) {
        return (d * 1000) + timeOffset;
    });

    endTimes = endTimes.map(function(d) {
        return (d * 1000) + timeOffset;
    });

    return {startTimes: startTimes, endTimes: endTimes};
}

// Create all longitudinal plots for all groups of metrics for a particular trial type
function allGroupsLongitudinalPlots(root, trials) {
    if (!trials || !trials.length) return;

    var defs = getDefinitionsForTrial(trials[0]);
    defs = getMetricDefClass(defs.metrics, "Measure");
    var metricGroups = metricsDefGroups(defs);

    // Build plots for all the groups of metrics
    var plotDiv = root.append("div").attr("class", "longitudinal-plot-group");
    plotDiv.append("h3").text("Longitudinal plots: " + trials[0].testDefinition.testName);

    for (var group in metricGroups) {
        var div = plotDiv.append("div").attr("class", "longitudinal-plot-group");
        groupLongitudinalPlots(div, getMetricDefGroup(defs, group),
                               group, trials);
    }
}

function groupLongitudinalPlots(root, defs, group, trials) {
    // Get the metrics we want out of every group and build some
    // objects
    var plotData = [];

    var trialData = trials.map(function(d) {
        return { date: d.date, condition: d.testDefinition.conditionName,
                 metrics: d.metrics.map( function(e) { return e; }) };
    });
    
    trialData.forEach(function(d) {
        d.metrics = d3.nest()
            .key(function(e) { return e.metricName; })
            .map(d.metrics, d3.map);
    });

    defs.forEach(function(d) {
        var metricData = { name: d.metricName, meanNorm: d.meanNorm,
                           sdNorm: d.sdNorm, units: d.units,
                           values: [] };
        trialData.forEach(function(e) {
            if (e.metrics[d.metricName]) {
                metricData.values
                    .push({ date: e.date, condition: e.condition,
                           value: e.metrics[d.metricName][0].mean });
            }
        });
        plotData.push(metricData);
    });

    // Build a series of plots for every metric in the group
    root.append("h4").text(group);
    plotData.forEach(function(d) {
        var plot = longitudinalPlot().width(200).height(200)
            .x(function(d) { return d.date; })
            .y(function(d) { return d.value; })
            .extra(function(d) { return d.condition; });

        var plotDiv = root.append("div").attr("id", "longitudinal-plot");
        plot.normMean(d.meanNorm).normStdDev(d.sdNorm)
            .xAxisLabel("Time/Date").yAxisLabel(d.units)
            .title(d.name);
        plotDiv.datum(d.values).call(plot);
        
    });

}

// Create the content and handlers for the navigation tree
function navigationTree(selection, data) {
    var tree = d3.layout.tree();
    var rootData = data;
    rootData.x0 = 0;
    rootData.y0 = 0;

    // Get the width from the current width of the div
    var width = 256; //selection.attr("width");
    var height = 1600;
    var barHeight = 20;
    var barWidth = width * 0.8;
    var duration = 100;
    var i = 0;
    var iconSize = 16;          // Assume icon is square
    var iconPad = 2;

    // The currently selected node
    var currNode = tree.nodes(data)[0];

    tree = tree.size([height, 50]);

    var navSvg = selection.append("svg").attr("id", "nav-tree-svg")
        .attr("width", width).attr("height", height);
    var navG = navSvg.append("g").attr("id", "nav-tree-main-g")
        .attr("transform", "translate(20,30)");

    collapseAll(rootData);
    navTreeUpdate(rootData);

    // Create the breadcrumbs
    breadcrumbs(currNode);

    function navTreeUpdate(root) {
        var nodes = tree.nodes(rootData);
        
        // Compute the "layout".
        nodes.forEach(function(n, i) {
            n.x = i * barHeight;
        });
        
        // Update the nodes…
        var node = navG.selectAll("g.node")
            .data(nodes, function(d) { return d.id || (d.id = ++i); });
        
        var nodeEnter = node.enter().append("svg:g")
            .attr("class", "node")
            .attr("transform", function() { return "translate(" +
                                            root.y0 + "," + root.x0 + ")"; })
            .style("opacity", 1e-6);

        // Enter any new nodes at the parent's previous position.
        nodeEnter.each( function(d) {
            if (d.children || d._children) {
                var arrow = d3.select(this).append("path")
                    .attr("class", "arrow");
                arrow.attr("d", d3.svg.symbol().type("triangle-down").size(36));
                arrow.attr("transform", function(d) {
                    if (d.children)
                        return "translate(-10,0) rotate(0)";
                    return "translate(-10,0) rotate(-90)"; });
                arrow.on("click", function() { return arrowClick(arrow, d); });
            }
        });
        
        nodeEnter.each( function(d) {
            var icon = d3.select(this).append("svg:image");
            icon.attr("x", -iconPad).attr("y", (-barHeight / 2) + iconPad)
                .attr("height", iconSize).attr("width", iconSize)
                .style("padding", 2);
            var iconPath = "/icons/folder.gif";
            if (d.type === "STUDY")
                iconPath = "/icons/magnifying_glass.png";
            else if (d.type === "STUDY_SUBJECT")
                iconPath = "/icons/people.png";
            else if (d.type === "TRIAL")
                iconPath = "/icons/disk.png";
            else if (d.type === "SESSION")
                iconPath = "/icons/clock.png";
            
            icon.attr("xlink:href", iconPath);
        });

        nodeEnter.append("svg:rect")
            .attr("y", -barHeight / 2)
            .attr("x", iconSize + iconPad)
            .attr("height", barHeight)
            .attr("width", barWidth *0.8)
            .style("fill", color)
            .on("click", function(d) { return click(d, true); });
        
        nodeEnter.append("svg:text")
            .attr("dy", 3.5)
            .attr("dx", 5.5 + iconSize + iconPad)
            .text(function(d) { return d.name; });
        
        // Transition nodes to their new position.
        nodeEnter.transition()
            .duration(duration)
            .attr("transform", function(d) { return "translate(" + d.y + "," + d.x + ")"; })
            .style("opacity", 1);
        
        node.transition()
            .duration(duration)
            .attr("transform", function(d) { return "translate(" + d.y + "," + d.x + ")"; })
            .style("opacity", 1)
            .select("rect")
            .style("fill", color);
        
        // Transition exiting nodes to the parent's new position.
        node.exit().transition()
            .duration(duration)
            .attr("transform", function() {
                return "translate(" + root.y + "," + root.x + ")"; })
            .style("opacity", 1e-6)
            .remove();
        
        // Stash the old positions for transition.
        nodes.forEach(function(d) {
            d.x0 = d.x;
            d.y0 = d.y;
        });
    }

    // Toggle children on click.
    function click(d, updateBreadcrumbs) {
        if (currNode === d) {
            var arrow = d3.select(d3.select(this).node().parentNode)
                .select("path");
            if (d.children) {
                arrow.attr("transform", "translate(-10,0) rotate(-90)");
                d._children = d.children;
                d.children = null;
            } else {
                arrow.attr("transform", "translate(-10,0) rotate(0)");
                d.children = d._children;
                d._children = null;
            }
        }
        if (updateBreadcrumbs === true) breadcrumbs(d);
        currNode = d;
        displayReport(d);
        navTreeUpdate(d);
    }

    // TODO: Rework colors to give correct terminal color to nodes
    function color(d) {
        if (d === currNode) return "#e6c71e";
        return d._children ? "#3182bd" : d.children ? "#c6dbef" : "#fd8d3c";
    }

    // Collapse all the given nodes children
    function collapseAll(node) {
        var nodes = tree.nodes(node);
        nodes.forEach( function(d) {
            if (d.children) {
                d._children = d.children;
                d.children = null;
            }
        });
        navTreeUpdate(node);
    }

    function arrowClick(arrow, node) {
        if (node.children) {
            arrow.attr("transform", "translate(-10,0) rotate(-90)");
            node._children = node.children;
            node.children = null;
        } else {
            arrow.attr("transform", "translate(-10,0) rotate(0)");
            node.children = node._children;
            node._children = null;
        }
        navTreeUpdate(node);
    }


    navTreeUpdate.currentNode = function(d) {
        if (!arguments.length) return currNode;
        currNode = d;
        return navTreeUpdate;
    };

    navTreeUpdate.expandNode = function(d) {
        if (!arguments.length) return currNode;
        currNode = d;

        // Expand all the parent nodes of the selected one
        var node = currNode;
        while (node) {
            if (node._children) {
                node.children = node._children;
                node._children = null;
            }
            node = node.parent;
        }
                
        click(d, true);
        return navTreeUpdate;
    };

    navTreeUpdate.selectNode = function(d) {
        if (!arguments.length) return currNode;
        click(d, true);
        return navTreeUpdate;
    };

    navTreeUpdate.breadcrumbSelectNode = function(d) {
        if (!arguments.length) return currNode;
        click(d, false);
        return navTreeUpdate;
    };

    return navTreeUpdate;
}

function displayReport(dataToReport) {

    var reportDiv = d3.select("#report");

    if (dataToReport.type === "STUDIES_NODE")
        studiesNodeReport(reportDiv, dataToReport);
    else if (dataToReport.type === "STUDY")
        studyReport(reportDiv, dataToReport);
    else if (dataToReport.type === "TRIALS_LIST")
        studyTrialsReport(reportDiv, dataToReport);
    else if (dataToReport.type === "STUDY_SUBJECT")
        studySubjectReport(reportDiv, dataToReport);
    else if (dataToReport.type === "SESSION")
        sessionReport(reportDiv, dataToReport);
    else
        singleTrialReport(reportDiv, dataToReport);
}

function breadcrumbs(currNode) {
    var height = 20;
    var width = 150;
    var duration = 200;
    var node = currNode;

    var crumbs = [];
    crumbs.push(node);
    while (node.parent) {
        node = node.parent;
        crumbs.unshift(node);
    }

    var svgArea = d3.select("#breadcrumbs").selectAll("svg")
        .data([currNode])
        .attr("height", height * 2)
        .attr("x", 0).attr("y", 0);
    
    var svg = svgArea.enter().append("svg");
    var gradient = svg.append("defs")
        .append("linearGradient")
        .attr("id", "crumb-gradient")
        .attr("x1", "0%").attr("y1", "0%")
        .attr("x2", "0%").attr("y2", "100%");
    gradient.append("stop")
        .attr("offset", "0%").attr("stop-color", "#5787c0").attr("stop-opacity", "1");
    gradient.append("stop")
        .attr("offset", "100%").attr("stop-color", "#a8c1de").attr("stop-opacity", "1");
    
    var crumb = svgArea.selectAll("g.crumb")
        .data(crumbs, function(d) { return d.id; });

    var crumbEnter = crumb.enter()
        .append("svg:g")
        .attr("class", "crumb")
        .attr("transform", function() { return "translate(0,0)"; })
        .style("opacity", 1e-6);

    crumbEnter.append("svg:text")
        .attr("dx", 5.5)
        .attr("dy", height - (height / 4))
        .text(function(d) { return d.name; });

    crumbEnter.each( function(d, i) {
        var g = d3.select(this);
        g.attr("transform", "translate(0,15)").style("opacity", 1e-6);
        var box = g.select("text").node().getBBox();
        if (i > 0)
            d.nextCrumbOffset = crumbs[i - 1].nextCrumbOffset + box.width + 20;
        else
            d.nextCrumbOffset = box.width + 20;
        g.insert("svg:path", "text")
            .style("fill", "url(#crumb-gradient)")
            .attr("d", crumbShape(box.width))
            .on("click", function(d) { navTree.breadcrumbSelectNode(d); });
    });

    crumbEnter.transition()
        .duration(duration)
        .attr("transform",
              function(d, i) {
                  var offset = 0;
                  if (i > 0)
                      offset = crumbs[i - 1].nextCrumbOffset;
                  return "translate(" + offset + ",15)"; })
        .style("opacity", 1);

    crumb.exit().transition()
        .duration(duration)
        .attr("transform", function(d, i) { return "translate(" +
                                            ((i - 1) * width) + ",15)"; })
        .style("opacity", 1e-6)
        .remove();

    function crumbShape(textWidth) {
        var path = "M 0 0 L " + (textWidth + 5) + " 0 L " + (textWidth + 20) + " " +
            (height / 2) +
            " L " + (textWidth + 5) + " " + height +
            " L 0 " + height + " Z";
        return path;
    }

}
